import argparse
import json
import logging
import math
import sys
import time
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.amp
import torch.nn.functional as F
import torchinfo
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torch.utils.tensorboard import SummaryWriter
from torchvision.datasets import FakeData
from torchvision.datasets import ImageFolder
from torchvision.datasets.folder import pil_loader  # Slower but Handles external dataset quirks better
from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import training_cli
from birder.common import training_utils
from birder.common.lib import format_duration
from birder.common.lib import get_network_name
from birder.common.lib import set_random_seeds
from birder.conf import settings
from birder.data.dataloader.webdataset import make_wds_loader
from birder.data.datasets.directory import HierarchicalImageFolder
from birder.data.datasets.directory import tv_loader
from birder.data.datasets.webdataset import make_wds_dataset
from birder.data.datasets.webdataset import prepare_wds_args
from birder.data.datasets.webdataset import wds_args_from_info
from birder.data.transforms.classification import get_mixup_cutmix
from birder.data.transforms.classification import get_rgb_stats
from birder.data.transforms.classification import inference_preset
from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import get_signature

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def train(args: argparse.Namespace) -> None:
    #
    # Initialize
    #
    training_utils.init_distributed_mode(args)
    logger.info(f"Starting training, birder version: {birder.__version__}, pytorch version: {torch.__version__}")
    training_utils.log_git_info()

    if args.size is None:
        args.size = registry.get_default_size(args.network)

    logger.info(f"Using size={args.size}")

    if args.cpu is True:
        device = torch.device("cpu")
        device_id = 0
    else:
        device = torch.device("cuda")
        device_id = torch.cuda.current_device()

    if args.use_deterministic_algorithms is True:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
    else:
        torch.backends.cudnn.benchmark = True

    if args.seed is not None:
        set_random_seeds(args.seed)

    if args.non_interactive is True or training_utils.is_local_primary(args) is False:
        disable_tqdm = True
    elif sys.stderr.isatty() is False:
        disable_tqdm = True
    else:
        disable_tqdm = False

    # Enable or disable the autograd anomaly detection
    torch.autograd.set_detect_anomaly(args.grad_anomaly_detection)

    #
    # Data
    #
    rgb_stats = get_rgb_stats(args.rgb_mode, args.rgb_mean, args.rgb_std)
    logger.debug(f"Using RGB stats: {rgb_stats}")

    training_transform = training_utils.get_training_transform(args)
    val_transform = inference_preset(args.size, rgb_stats, 1.0)
    if args.use_fake_data is True:
        logger.warning("Using fake data")
        training_dataset = FakeData(10000, (args.channels, *args.size), num_classes=10, transform=training_transform)
        validation_dataset = FakeData(1000, (args.channels, *args.size), num_classes=10, transform=val_transform)
        class_to_idx = {str(i): i for i in range(10)}

    elif args.wds is True:
        training_wds_path: str | list[str]
        val_wds_path: str | list[str]
        if args.wds_info is not None:
            (training_wds_path, training_size) = wds_args_from_info(args.wds_info, args.wds_training_split)
            (val_wds_path, val_size) = wds_args_from_info(args.wds_info, args.wds_val_split)
            if args.wds_train_size is not None:
                training_size = args.wds_train_size
            if args.wds_val_size is not None:
                val_size = args.wds_val_size
        else:
            (training_wds_path, training_size) = prepare_wds_args(args.data_path, args.wds_train_size, device)
            (val_wds_path, val_size) = prepare_wds_args(args.val_path, args.wds_val_size, device)

        training_dataset = make_wds_dataset(
            training_wds_path,
            dataset_size=training_size,
            shuffle=True,
            samples_names=False,
            transform=training_transform,
            img_loader=args.img_loader,
            cache_dir=args.wds_cache_dir,
        )
        validation_dataset = make_wds_dataset(
            val_wds_path,
            dataset_size=val_size,
            shuffle=False,
            samples_names=False,
            transform=val_transform,
            img_loader=args.img_loader,
            cache_dir=args.wds_cache_dir,
        )

        class_to_idx = fs_ops.read_class_file(args.wds_class_file)

    else:
        if args.hierarchical is True:
            dataset_cls = HierarchicalImageFolder
        else:
            dataset_cls = ImageFolder

        training_dataset = dataset_cls(
            args.data_path, transform=training_transform, loader=pil_loader if args.img_loader == "pil" else tv_loader
        )
        validation_dataset = dataset_cls(
            args.val_path,
            transform=val_transform,
            loader=pil_loader if args.img_loader == "pil" else tv_loader,
            allow_empty=True,
        )
        assert training_dataset.class_to_idx == validation_dataset.class_to_idx
        class_to_idx = training_dataset.class_to_idx

    assert args.model_ema is False or args.model_ema_steps <= len(training_dataset) / args.batch_size

    logger.info(f"Using device {device}:{device_id}")
    logger.info(f"Training on {len(training_dataset):,} samples")
    logger.info(f"Validating on {len(validation_dataset):,} samples")

    num_outputs = len(class_to_idx)
    batch_size: int = args.batch_size
    model_ema_steps: int = args.model_ema_steps * args.grad_accum_steps
    logger.debug(f"Effective batch size = {args.batch_size * args.grad_accum_steps * args.world_size}")

    # Set data iterators
    if args.mixup_alpha is not None or args.cutmix is True:
        logger.debug("Mixup / cutmix collate activated")
        t = get_mixup_cutmix(args.mixup_alpha, num_outputs, args.cutmix)

        def collate_fn(batch: Any) -> Any:
            return t(*default_collate(batch))

    else:
        collate_fn = None  # type: ignore

    # Data loaders and samplers
    (train_sampler, validation_sampler) = training_utils.get_samplers(args, training_dataset, validation_dataset)

    if args.wds is True:
        training_loader = make_wds_loader(
            training_dataset,
            batch_size,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=collate_fn,
            world_size=args.world_size,
            pin_memory=True,
            drop_last=args.drop_last,
            shuffle=args.wds_extra_shuffle,
        )

        validation_loader = make_wds_loader(
            validation_dataset,
            batch_size,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=None,
            world_size=args.world_size,
            pin_memory=True,
        )

    else:
        training_loader = DataLoader(
            training_dataset,
            batch_size=batch_size,
            sampler=train_sampler,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            collate_fn=collate_fn,
            pin_memory=True,
            drop_last=args.drop_last,
        )
        validation_loader = DataLoader(
            validation_dataset,
            batch_size=batch_size,
            sampler=validation_sampler,
            num_workers=args.num_workers,
            prefetch_factor=args.prefetch_factor,
            pin_memory=True,
        )

    last_batch_idx = len(training_loader) - 1
    begin_epoch = 1
    epochs = args.epochs + 1
    if args.stop_epoch is None:
        args.stop_epoch = epochs
    else:
        args.stop_epoch += 1

    #
    # Initialize network
    #
    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    sample_shape = (batch_size, args.channels, *args.size)  # B, C, H, W
    network_name = get_network_name(args.network, tag=args.tag)

    if args.resume_epoch is not None:
        begin_epoch = args.resume_epoch + 1
        (net, class_to_idx_saved, training_states) = fs_ops.load_checkpoint(
            device,
            args.network,
            config=args.model_config,
            tag=args.tag,
            epoch=args.resume_epoch,
            new_size=args.size,
            strict=not args.non_strict_weights,
        )
        if args.reset_head is True:
            net.reset_classifier(len(class_to_idx))
        else:
            assert class_to_idx == class_to_idx_saved

    elif args.pretrained is True:
        fs_ops.download_model_by_weights(network_name, progress_bar=training_utils.is_local_primary(args))
        (net, class_to_idx_saved, training_states) = fs_ops.load_checkpoint(
            device,
            args.network,
            config=args.model_config,
            tag=args.tag,
            epoch=None,
            new_size=args.size,
            strict=not args.non_strict_weights,
        )
        if args.reset_head is True:
            net.reset_classifier(len(class_to_idx))
        else:
            assert class_to_idx == class_to_idx_saved

    else:
        net = registry.net_factory(args.network, sample_shape[1], num_outputs, config=args.model_config, size=args.size)
        training_states = fs_ops.TrainingStates.empty()

    net.to(device, dtype=model_dtype)
    if args.freeze_body is True:
        net.freeze(freeze_classifier=False, unfreeze_features=args.unfreeze_features)
    elif args.freeze_stages is not None:
        net.freeze_stages(up_to_stage=args.freeze_stages)

    if args.freeze_bn is True:
        net = training_utils.freeze_batchnorm2d(net)
    elif args.sync_bn is True and args.distributed is True:
        net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net)

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    # Compile network
    if args.compile is True:
        net = torch.compile(net)

    #
    # Loss criteria, optimizer, learning rate scheduler and training parameter groups
    #

    # Training parameter groups
    custom_keys_weight_decay = training_utils.get_wd_custom_keys(args)
    parameters = training_utils.optimizer_parameter_groups(
        net,
        args.wd,
        norm_weight_decay=args.norm_wd,
        custom_keys_weight_decay=custom_keys_weight_decay,
        layer_decay=args.layer_decay,
        layer_decay_min_scale=args.layer_decay_min_scale,
        layer_decay_no_opt_scale=args.layer_decay_no_opt_scale,
        bias_lr=args.bias_lr,
    )

    # Loss criteria
    if args.bce_loss is True:
        criterion = torch.nn.BCEWithLogitsLoss()
    else:
        criterion = torch.nn.CrossEntropyLoss(label_smoothing=args.smoothing_alpha)

    # Learning rate scaling
    lr = training_utils.scale_lr(args)
    grad_accum_steps: int = args.grad_accum_steps

    if args.lr_scheduler_update == "epoch":
        iter_update = False
        iters_per_epoch = 1
    elif args.lr_scheduler_update == "iter":
        iter_update = True
        iters_per_epoch = math.ceil(len(training_loader) / grad_accum_steps)
    else:
        raise ValueError("Unsupported lr_scheduler_update")

    # Optimizer and learning rate scheduler
    optimizer = training_utils.get_optimizer(parameters, lr, args)
    scheduler = training_utils.get_scheduler(optimizer, iters_per_epoch, args)
    if args.compile_opt is True:
        optimizer.step = torch.compile(optimizer.step, fullgraph=False)

    # Gradient scaler and AMP related tasks
    (scaler, amp_dtype) = training_utils.get_amp_scaler(args.amp, args.amp_dtype)

    # Load states
    if args.load_states is True:
        optimizer.load_state_dict(training_states.optimizer_state)
        scheduler.load_state_dict(training_states.scheduler_state)
        if scaler is not None:
            scaler.load_state_dict(training_states.scaler_state)

    elif args.load_scheduler is True:
        scheduler.load_state_dict(training_states.scheduler_state)
        last_lrs = scheduler.get_last_lr()
        for g, last_lr in zip(optimizer.param_groups, last_lrs):
            g["lr"] = last_lr

    last_lr = max(scheduler.get_last_lr())
    if args.plot_lr is True:
        logger.info("Fast forwarding scheduler...")
        optimizer.step()
        lrs = []
        for _ in range(begin_epoch, epochs):
            for _ in range(iters_per_epoch):
                lrs.append(max(scheduler.get_last_lr()))
                scheduler.step()

        plt.plot(np.linspace(begin_epoch, epochs, iters_per_epoch * (epochs - begin_epoch), endpoint=False), lrs)
        plt.show()
        raise SystemExit(0)

    #
    # Distributed (DDP) and Model EMA
    #
    if args.model_ema_warmup is not None:
        ema_warmup_epochs = args.model_ema_warmup
    elif args.warmup_epochs is not None:
        ema_warmup_epochs = args.warmup_epochs
    elif args.warmup_iters is not None:
        ema_warmup_epochs = args.warmup_iters // iters_per_epoch
    else:
        ema_warmup_epochs = 0

    logger.debug(f"EMA warmup epochs = {ema_warmup_epochs}")
    net_without_ddp = net
    if args.distributed is True:
        net = torch.nn.parallel.DistributedDataParallel(
            net, device_ids=[args.local_rank], find_unused_parameters=args.find_unused_parameters
        )
        net_without_ddp = net.module

    if args.model_ema is True:
        model_base = net_without_ddp  # Original model without DDP wrapper, will be saved as training state
        model_ema = training_utils.ema_model(args, net_without_ddp, device=device)
        if args.load_states is True and training_states.ema_model_state is not None:
            logger.info("Setting model EMA weights...")
            if args.compile is True and hasattr(model_ema.module, "_orig_mod") is True:
                model_ema.module._orig_mod.load_state_dict(  # pylint: disable=protected-access
                    training_states.ema_model_state
                )
            else:
                model_ema.module.load_state_dict(training_states.ema_model_state)

            model_ema.n_averaged += 1  # pylint:disable=no-member

        model_to_save = model_ema.module  # Save EMA model weights as default weights
        eval_model = model_ema  # Use EMA for evaluation

    else:
        model_base = None
        model_to_save = net_without_ddp
        eval_model = net

    if args.compile is True and hasattr(model_to_save, "_orig_mod") is True:
        model_to_save = model_to_save._orig_mod  # pylint: disable=protected-access
    if args.compile is True and hasattr(model_base, "_orig_mod") is True:
        model_base = model_base._orig_mod  # type: ignore[union-attr] # pylint: disable=protected-access

    #
    # Misc
    #

    # Print network summary
    net_for_info = net_without_ddp
    if args.compile is True and hasattr(net_without_ddp, "_orig_mod") is True:
        net_for_info = net_without_ddp._orig_mod  # pylint: disable=protected-access

    if args.no_summary is False:
        summary = torchinfo.summary(
            net_for_info,
            device=device,
            input_size=sample_shape,
            dtypes=[model_dtype],
            col_names=["input_size", "output_size", "kernel_size", "num_params"],
            depth=4,
            verbose=0,
        )
        if training_utils.is_global_primary(args) is True:
            # Write to stderr, same as all the logs
            print(summary, file=sys.stderr)

    # Training logs
    training_log_path = training_utils.training_log_path(network_name, device, args.experiment)
    logger.info(f"Logging training run at {training_log_path}")
    summary_writer = SummaryWriter(training_log_path)

    signature = get_signature(input_shape=sample_shape, num_outputs=num_outputs)
    file_handler: logging.Handler = logging.NullHandler()
    if training_utils.is_local_primary(args) is True:
        with torch.no_grad():
            summary_writer.add_graph(net_for_info, torch.rand(sample_shape, device=device, dtype=model_dtype))

        summary_writer.flush()
        fs_ops.write_config(network_name, net_for_info, signature=signature, rgb_stats=rgb_stats)
        file_handler = training_utils.setup_file_logging(training_log_path.joinpath("training.log"))
        with open(training_log_path.joinpath("training_args.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "birder_version": birder.__version__,
                    "pytorch_version": torch.__version__,
                    "cmdline": " ".join(sys.argv),
                    **vars(args),
                },
                handle,
                indent=2,
            )

        with open(training_log_path.joinpath("training_data.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "training_samples": len(training_dataset),
                    "validation_samples": len(validation_dataset),
                    "classes": list(class_to_idx.keys()),
                },
                handle,
                indent=2,
            )

    #
    # Training loop
    #
    logger.info(f"Starting training with learning rate of {last_lr}")
    for epoch in range(begin_epoch, args.stop_epoch):
        tic = time.time()
        net.train()
        running_loss = training_utils.SmoothedValue(window_size=64)
        running_val_loss = training_utils.SmoothedValue()
        train_accuracy = training_utils.SmoothedValue(window_size=64)
        val_accuracy = training_utils.SmoothedValue()

        if args.distributed is True:
            train_sampler.set_epoch(epoch)

        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=len(training_dataset),
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )

        # Zero the parameter gradients
        optimizer.zero_grad()

        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0
        for i, (inputs, targets) in enumerate(training_loader):
            inputs = inputs.to(device, dtype=model_dtype, non_blocking=True)
            targets = targets.to(device, non_blocking=True)
            loss_targets = targets
            if args.bce_loss is True:
                if loss_targets.ndim == 1:
                    loss_targets = F.one_hot(loss_targets, num_classes=num_outputs)  # pylint: disable=not-callable

                loss_targets = loss_targets.gt(args.bce_threshold).to(dtype=inputs.dtype)

            optimizer_update = (i == last_batch_idx) or ((i + 1) % grad_accum_steps == 0)

            # Forward, backward and optimize
            with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                outputs = net(inputs)
                loss = criterion(outputs, loss_targets)

            if scaler is not None:
                scaler.scale(loss).backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    scaler.step(optimizer)
                    scaler.update()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            else:
                loss.backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    optimizer.step()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            # Exponential moving average
            if args.model_ema is True and i % model_ema_steps == 0:
                model_ema.update_parameters(net)
                if epoch <= ema_warmup_epochs:
                    # Reset ema buffer to keep copying weights during warmup period
                    model_ema.n_averaged.fill_(0)  # pylint: disable=no-member

            # Statistics
            running_loss.update(loss.detach())
            if targets.ndim == 2:
                targets = targets.argmax(dim=1)

            train_accuracy.update(training_utils.accuracy(targets, outputs.detach()))

            # Write statistics
            if (i == last_batch_idx) or (i + 1) % args.log_interval == 0:
                time_now = time.time()
                time_cost = time_now - start_time
                steps_processed_in_interval = i - last_idx
                rate = steps_processed_in_interval * (batch_size * args.world_size) / time_cost

                avg_time_per_step = time_cost / steps_processed_in_interval
                remaining_steps_in_epoch = last_batch_idx - i
                estimated_time_to_finish_epoch = remaining_steps_in_epoch * avg_time_per_step

                start_time = time_now
                last_idx = i
                cur_lr = max(scheduler.get_last_lr())

                running_loss.synchronize_between_processes(device)
                train_accuracy.synchronize_between_processes(device)
                with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
                    log.info(
                        f"[Trn] Epoch {epoch}/{epochs-1}, step {i+1}/{last_batch_idx+1}  "
                        f"Loss: {running_loss.avg:.4f}  "
                        f"Elapsed: {format_duration(time_now-epoch_start)}  "
                        f"ETA: {format_duration(estimated_time_to_finish_epoch)}  "
                        f"T: {time_cost:.1f}s  "
                        f"R: {rate:.1f} samples/s  "
                        f"LR: {cur_lr:.4e}"
                    )
                    log.info(
                        f"[Trn] Epoch {epoch}/{epochs-1}, step {i+1}/{last_batch_idx+1}  "
                        f"Accuracy: {train_accuracy.avg:.4f}"
                    )

                if training_utils.is_local_primary(args) is True:
                    summary_writer.add_scalars(
                        "loss",
                        {"training": running_loss.avg},
                        ((epoch - 1) * len(training_dataset)) + (i * batch_size * args.world_size),
                    )
                    summary_writer.add_scalars(
                        "performance",
                        {"training_accuracy": train_accuracy.avg},
                        ((epoch - 1) * len(training_dataset)) + (i * batch_size * args.world_size),
                    )

            # Update progress bar
            progress.update(n=batch_size * args.world_size)

        progress.close()

        # Epoch training metrics
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} training_loss: {running_loss.global_avg:.4f}")
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} training_accuracy: {train_accuracy.global_avg:.4f}")

        # Validation
        eval_model.eval()
        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=len(validation_dataset),
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )
        with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
            log.info(f"[Val] Starting validation for epoch {epoch}/{epochs-1}...")

        epoch_start = time.time()
        with torch.inference_mode():
            for inputs, targets in validation_loader:
                inputs = inputs.to(device, dtype=model_dtype, non_blocking=True)
                targets = targets.to(device, non_blocking=True)
                loss_targets = targets
                if args.bce_loss is True:
                    loss_targets = F.one_hot(loss_targets, num_classes=num_outputs)  # pylint: disable=not-callable
                    loss_targets = loss_targets.to(dtype=inputs.dtype)

                with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                    outputs = eval_model(inputs)
                    val_loss = criterion(outputs, loss_targets)

                # Statistics
                running_val_loss.update(val_loss.detach())
                val_accuracy.update(training_utils.accuracy(targets, outputs), n=outputs.size(0))

                # Update progress bar
                progress.update(n=batch_size * args.world_size)

        time_now = time.time()
        rate = len(validation_dataset) / (time_now - epoch_start)
        with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
            log.info(
                f"[Val] Epoch {epoch}/{epochs-1} "
                f"Elapsed: {format_duration(time_now-epoch_start)}  "
                f"R: {rate:.1f} samples/s"
            )

        progress.close()

        running_val_loss.synchronize_between_processes(device)
        val_accuracy.synchronize_between_processes(device)
        epoch_val_loss = running_val_loss.global_avg
        epoch_val_accuracy = val_accuracy.global_avg

        # Write statistics
        if training_utils.is_local_primary(args) is True:
            summary_writer.add_scalars("loss", {"validation": epoch_val_loss}, epoch * len(training_dataset))
            summary_writer.add_scalars(
                "performance", {"validation_accuracy": epoch_val_accuracy}, epoch * len(training_dataset)
            )

        # Epoch validation metrics
        logger.info(f"[Val] Epoch {epoch}/{epochs-1} validation_loss: {epoch_val_loss:.4f}")
        logger.info(f"[Val] Epoch {epoch}/{epochs-1} validation_accuracy: {epoch_val_accuracy:.4f}")

        # Learning rate scheduler update
        if iter_update is False:
            scheduler.step()
        if last_lr != max(scheduler.get_last_lr()):
            last_lr = max(scheduler.get_last_lr())
            logger.info(f"Updated learning rate to: {last_lr}")

        if training_utils.is_local_primary(args) is True:
            # Checkpoint model
            if epoch % args.save_frequency == 0:
                fs_ops.checkpoint_model(
                    network_name,
                    epoch,
                    model_to_save,
                    signature,
                    class_to_idx,
                    rgb_stats,
                    optimizer,
                    scheduler,
                    scaler,
                    model_base,
                )
                if args.keep_last is not None:
                    fs_ops.clean_checkpoints(network_name, args.keep_last)

        # Epoch timing
        toc = time.time()
        logger.info(f"Total time: {format_duration(toc - tic)}")
        logger.info("---")

        # Reset counters
        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0

    # Save model hyperparameters with metrics
    if training_utils.is_local_primary(args) is True:
        # Replace list/dict based args
        if args.opt_betas is not None:
            for idx, beta in enumerate(args.opt_betas):
                setattr(args, f"opt_betas_{idx}", beta)

            del args.opt_betas

        if args.lr_steps is not None:
            args.lr_steps = json.dumps(args.lr_steps)
        if args.model_config is not None:
            args.model_config = json.dumps(args.model_config)
        if args.size is not None:
            args.size = json.dumps(args.size)

        # Save all args
        summary_writer.add_hparams(
            {**vars(args), "training_samples": len(training_dataset)},
            {
                "hparam/acc": train_accuracy.global_avg,
                "hparam/val_acc": val_accuracy.global_avg,
            },
        )

    summary_writer.close()

    # Checkpoint model
    if training_utils.is_local_primary(args) is True:
        fs_ops.checkpoint_model(
            network_name,
            epoch,
            model_to_save,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer,
            scheduler,
            scaler,
            model_base,
        )

    training_utils.shutdown_distributed_mode(args)


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Train classification model",
        epilog=(
            "Usage examples\n"
            "==============\n"
            "Training on image directory in the default location can be as simple as:\n"
            "python train.py --network densenet_161 --lr-scheduler cosine --batch-size 64 --smoothing-alpha 0.1\n"
            "\n"
            "A more complicated ImageNet example on 2 GPUs, using a remote webdataset:\n"
            "torchrun --nproc_per_node=2 train.py \\\n"
            "    --network resnet_v2_50 \\\n"
            "    --tag imagenet1k \\\n"
            "    --opt lamb \\\n"
            "    --lr 0.005 \\\n"
            "    --lr-scheduler cosine \\\n"
            "    --lr-cosine-min 1e-7 \\\n"
            "    --warmup-epochs 5 \\\n"
            "    --epochs 300 \\\n"
            "    --wd 0.02 \\\n"
            "    --grad-accum-steps 4 \\\n"
            "    --mixup-alpha 0.1 \\\n"
            "    --cutmix \\\n"
            "    --aug-type ra \\\n"
            "    --re-prob 0.25 \\\n"
            "    --rgb-mode imagenet \\\n"
            "    --bce-loss --bce-threshold 0.2 \\\n"
            "    --amp \\\n"
            "    --compile \\\n"
            "    --wds \\\n"
            "    --wds-class-file public_datasets_metadata/imagenet-1k-classes.txt \\\n"
            "    --wds-info https://huggingface.co/datasets/timm/imagenet-1k-wds/resolve/main/_info.json \\\n"
            "    --wds-training-split train\n"
            "\n"
            "The script is also available as module:\n"
            "torchrun --nproc_per_node=2 -m birder.scripts.train \\\n"
            "    --network regnet_y_8g \\\n"
            "    --lr 0.4 \\\n"
            "    --lr-scheduler cosine \\\n"
            "    --warmup-epochs 5 \\\n"
            "    --wd 0.00005 \\\n"
            "    --smoothing-alpha 0.1 \\\n"
            "    --mixup-alpha 0.2 \\\n"
            "    --cutmix \\\n"
            "    --aug-level 8 \\\n"
            "    --ra-sampler --ra-reps 2 \\\n"
            "    --amp --compile\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("-t", "--tag", type=str, help="add model tag")
    parser.add_argument("--reset-head", default=False, action="store_true", help="reset the classification head")
    parser.add_argument(
        "--freeze-body",
        default=False,
        action="store_true",
        help="freeze all layers of the model except the classification head",
    )
    parser.add_argument("--freeze-stages", type=int, help="number of model stages to freeze (for supported models)")
    parser.add_argument(
        "--unfreeze-features",
        default=False,
        action="store_true",
        help="unfreeze features layer (only relevant when freezing body)",
    )

    group = parser.add_argument_group("Loss parameters")
    group.add_argument("--smoothing-alpha", type=float, default=0.0, help="label smoothing alpha")
    group.add_argument("--bce-loss", default=False, action="store_true", help="enable BCE loss")
    group.add_argument("--bce-threshold", type=float, default=0.0, help="threshold for binarizing soft BCE targets")
    training_cli.add_optimization_args(parser)
    training_cli.add_lr_wd_args(parser)
    training_cli.add_lr_scheduler_args(parser)
    training_cli.add_training_schedule_args(parser)
    training_cli.add_input_args(parser)
    training_cli.add_data_aug_args(parser, mixup_cutmix=True)
    training_cli.add_ema_args(parser)
    training_cli.add_dataloader_args(parser, ra_sampler=True)
    training_cli.add_batch_norm_args(parser)
    training_cli.add_precision_args(parser)
    training_cli.add_compile_args(parser)
    training_cli.add_checkpoint_args(parser, default_save_frequency=5, pretrained=True)
    training_cli.add_distributed_args(parser)
    training_cli.add_logging_and_debug_args(parser)
    training_cli.add_training_data_args(parser)

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.data_path = str(args.data_path)
    args.val_path = str(args.val_path)
    args.size = cli.parse_size(args.size)

    # This will capture the common argument mistakes
    training_cli.common_args_validation(args)

    # Script specific checks
    if registry.exists(args.network, task=Task.IMAGE_CLASSIFICATION) is False:
        raise cli.ValidationError(f"--network {args.network} not supported, see list-models tool for available options")

    if args.smoothing_alpha < 0 or args.smoothing_alpha >= 0.5:
        raise cli.ValidationError(f"--smoothing-alpha must be in range of [0, 0.5), got {args.smoothing_alpha}")
    if args.bce_loss is True and args.smoothing_alpha != 0.0:
        raise cli.ValidationError("--bce-loss can only be used with --smoothing-alpha 0.0")

    if args.freeze_body is True and args.freeze_stages is not None:
        raise cli.ValidationError("--freeze-body cannot be used with --freeze-stages")
    if args.freeze_stages is not None and registry.exists(args.network, net_type=DetectorBackbone) is False:
        raise cli.ValidationError(
            "--freeze-stages only supported on detector backbone type networks, "
            "see list-models tool for available options"
        )


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.MODELS_DIR.exists() is False:
        logger.info(f"Creating {settings.MODELS_DIR} directory...")
        settings.MODELS_DIR.mkdir(parents=True, exist_ok=True)

    if args.wds_cache_dir is not None and Path(args.wds_cache_dir).exists() is False:
        logger.info(f"Creating {args.wds_cache_dir} directory...")
        Path(args.wds_cache_dir).mkdir(parents=True, exist_ok=True)

    train(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
EdgeNeXt, adapted from
https://github.com/mmaaz60/EdgeNeXt

Paper "EdgeNeXt: Efficiently Amalgamated CNN-Transformer Architecture for Mobile Vision Applications",
https://arxiv.org/abs/2206.10589
"""

# Reference license: MIT

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.xcit import PositionalEncodingFourier


class ConvBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        kernel_size: tuple[int, int],
        expand_ratio: int,
        layer_scale_init_value: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.dw_conv = nn.Conv2d(
            dim,
            dim_out,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2),
            groups=dim,
        )
        self.norm = nn.LayerNorm(dim_out, eps=1e-6)
        self.pw_conv1 = nn.Linear(dim_out, expand_ratio * dim_out)
        self.act = nn.GELU()
        self.pw_conv2 = nn.Linear(expand_ratio * dim_out, dim_out)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones(dim_out))
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.dw_conv(x)
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.pw_conv2(x)
        x = self.gamma * x
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)
        x = shortcut + self.drop_path(x)

        return x


class CrossCovarianceAttn(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(self.num_heads, 1, 1))

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, -1)
        qkv = qkv.permute(2, 0, 3, 4, 1)
        (q, k, v) = qkv.unbind(0)

        q = F.normalize(q, dim=-1) * self.temperature
        k = F.normalize(k, dim=-1)
        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=1.0
        )

        x = x.permute(0, 3, 1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class SDTA(nn.Module):
    def __init__(
        self,
        dim: int,
        drop_path: float,
        layer_scale_init_value: float,
        expand_ratio: int,
        use_pos_emb: bool,
        num_heads: int,
        qkv_bias: bool,
        attn_drop: float,
        drop: float,
        num_scales: int,
    ) -> None:
        super().__init__()
        width = max(int(math.ceil(dim / num_scales)), int(math.floor(dim // num_scales)))
        self.width = width
        self.num_scales = max(1, num_scales - 1)

        conv_list = []
        for _ in range(self.num_scales):
            conv_list.append(nn.Conv2d(width, width, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=width))

        self.conv_list = nn.ModuleList(conv_list)

        self.pos_embed = None
        if use_pos_emb is True:
            self.pos_embed = PositionalEncodingFourier(hidden_dim=32, dim=dim)

        self.norm_xca = nn.LayerNorm(dim, eps=1e-6)
        self.gamma_xca = nn.Parameter(layer_scale_init_value * torch.ones(dim))
        self.xca = CrossCovarianceAttn(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)

        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pw_conv1 = nn.Linear(dim, expand_ratio * dim)
        self.act = nn.GELU()
        self.pw_conv2 = nn.Linear(expand_ratio * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)))
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x

        # Replaced split with workaround from:
        # https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/edgenext.py
        spx = x.chunk(len(self.conv_list) + 1, dim=1)
        sp_out = []
        sp = spx[0]
        for i, conv in enumerate(self.conv_list):
            if i > 0:
                sp = sp + spx[i]

            sp = conv(sp)
            sp_out.append(sp)

        sp_out.append(spx[-1])
        x = torch.concat(sp_out, 1)

        # XCA
        (B, C, H, W) = x.shape
        x = x.reshape(B, C, H * W).permute(0, 2, 1)
        if self.pos_embed is not None:
            pos_encoding = self.pos_embed(B, H, W).reshape(B, -1, x.shape[1]).permute(0, 2, 1)
            x = x + pos_encoding

        x = x + self.drop_path(self.gamma_xca * self.xca(self.norm_xca(x)))
        x = x.reshape(B, H, W, C)

        # Inverted Bottleneck
        x = self.norm(x)
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.pw_conv2(x)
        x = self.gamma * x
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = shortcut + self.drop_path(x)

        return x


class EdgeNeXtStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        depth: int,
        num_global_blocks: int,
        num_heads: int,
        scales: int,
        kernel_size: tuple[int, int],
        expand_ratio: int,
        use_pos_emb: bool,
        layer_scale_init_value: float,
        drop_path_rates: list[float],
    ) -> None:
        super().__init__()
        self.grad_checkpointing = False

        if stride == 1:
            self.downsample = nn.Identity()
        else:
            self.downsample = nn.Sequential(
                LayerNorm2d(in_channels, eps=1e-6),
                nn.Conv2d(in_channels, out_channels, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
            )
            in_channels = out_channels

        stage_blocks = []
        for i in range(depth):
            if i < depth - num_global_blocks:
                stage_blocks.append(
                    ConvBlock(
                        dim=in_channels,
                        dim_out=out_channels,
                        kernel_size=kernel_size,
                        expand_ratio=expand_ratio,
                        layer_scale_init_value=layer_scale_init_value,
                        drop_path=drop_path_rates[i],
                    )
                )
            else:
                stage_blocks.append(
                    SDTA(
                        dim=in_channels,
                        num_scales=scales,
                        num_heads=num_heads,
                        expand_ratio=expand_ratio,
                        use_pos_emb=use_pos_emb,
                        layer_scale_init_value=layer_scale_init_value,
                        drop_path=drop_path_rates[i],
                        qkv_bias=True,
                        attn_drop=0.0,
                        drop=0.0,
                    )
                )
            in_channels = out_channels

        self.blocks = nn.Sequential(*stage_blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class EdgeNeXt(DetectorBackbone):
    default_size = (256, 256)

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        global_block_counts = [0, 1, 1, 1]
        d2_scales = [2, 2, 3, 4]
        expand_ratio = 4
        kernel_sizes = [(3, 3), (5, 5), (7, 7), (9, 9)]
        use_pos_emb = [False, True, False, False]
        layer_scale_init_value = 1e-6
        depths: list[int] = self.config["depths"]
        dims: list[int] = self.config["dims"]
        heads: list[int] = self.config["heads"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            dims[0],
            kernel_size=(4, 4),
            stride=(4, 4),
            padding=(0, 0),
            bias=True,
            norm_layer=LayerNorm2d,
            activation_layer=None,
        )

        current_stride = 4
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        in_channels = dims[0]
        for i, (depth, dim, head) in enumerate(zip(depths, dims, heads)):
            stride = 2 if current_stride == 2 or i > 0 else 1
            current_stride *= stride
            stages[f"stage{i+1}"] = EdgeNeXtStage(
                in_channels=in_channels,
                out_channels=dim,
                stride=stride,
                depth=depth,
                num_global_blocks=global_block_counts[i],
                num_heads=head,
                drop_path_rates=dpr[i],
                scales=d2_scales[i],
                expand_ratio=expand_ratio,
                kernel_size=kernel_sizes[i],
                use_pos_emb=use_pos_emb[i],
                layer_scale_init_value=layer_scale_init_value,
            )
            return_channels.append(dim)
            in_channels = dim

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(dims[-1], eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = dims[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "edgenext_xxs",
    EdgeNeXt,
    config={"depths": [2, 2, 6, 2], "dims": [24, 48, 88, 168], "heads": [4, 4, 4, 4], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "edgenext_xs",
    EdgeNeXt,
    config={"depths": [3, 3, 9, 3], "dims": [32, 64, 100, 192], "heads": [4, 4, 4, 4], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "edgenext_s",
    EdgeNeXt,
    config={"depths": [3, 3, 9, 3], "dims": [48, 96, 160, 304], "heads": [8, 8, 8, 8], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "edgenext_b",
    EdgeNeXt,
    config={"depths": [3, 3, 9, 3], "dims": [80, 160, 288, 584], "heads": [8, 8, 8, 8], "drop_path_rate": 0.1},
)

registry.register_weights(
    "edgenext_xxs_il-common",
    {
        "description": "EdgeNeXt extra extra small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 4.7,
                "sha256": "05f567111bebe05f3e88cb73aa665407009decf275836ec3c1c823b3725c6725",
            }
        },
        "net": {"network": "edgenext_xxs", "tag": "il-common"},
    },
)
registry.register_weights(
    "edgenext_xs_il-common",
    {
        "description": "EdgeNeXt extra small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 8.5,
                "sha256": "ceb0554f66dfaeb8cbba95137982756b1a5456f5359dda1657f9aab934ed7e0e",
            }
        },
        "net": {"network": "edgenext_xs", "tag": "il-common"},
    },
)
registry.register_weights(
    "edgenext_s_il-common",
    {
        "description": "EdgeNeXt small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 20.7,
                "sha256": "2855125a641f143b97fff4ef27ac1386a3058d1659b4e8e5d723952e788ce0ef",
            }
        },
        "net": {"network": "edgenext_s", "tag": "il-common"},
    },
)



"""
ViT Parallel, adapted from
https://github.com/facebookresearch/deit/blob/main/models_v2.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vision_transformer.py

Paper "Three things everyone should know about Vision Transformers", https://arxiv.org/abs/2203.09795
"""

# Reference license: Apache-2.0 (both)

import math
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerScale
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding


class Attention(nn.Module):
    def __init__(self, dim: int, num_heads: int, attn_drop: float, proj_drop: float) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class EncoderParallelBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: Optional[int],
        num_parallel: int,
        dropout: float,
        attention_dropout: float,
        drop_path: float,
        activation_layer: Callable[..., nn.Module],
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        if mlp_dim is None:
            mlp_dim = hidden_dim * 4

        # Attention blocks
        self.attn_blocks = nn.ModuleList()
        for _ in range(num_parallel):
            self.attn_blocks.append(
                nn.Sequential(
                    norm_layer(hidden_dim, eps=1e-6),
                    Attention(hidden_dim, num_heads, attn_drop=attention_dropout, proj_drop=dropout),
                    LayerScale(hidden_dim, layer_scale_init_value) if layer_scale_init_value else nn.Identity(),
                    StochasticDepth(drop_path, mode="row"),
                )
            )

        # MLP blocks
        self.mlp_blocks = nn.ModuleList()
        for _ in range(num_parallel):
            self.mlp_blocks.append(
                nn.Sequential(
                    norm_layer(hidden_dim, eps=1e-6),
                    MLP(
                        hidden_dim,
                        [mlp_dim, hidden_dim],
                        activation_layer=activation_layer,
                        inplace=None,
                        dropout=dropout,
                    ),
                    LayerScale(hidden_dim, layer_scale_init_value) if layer_scale_init_value else nn.Identity(),
                    StochasticDepth(drop_path, mode="row"),
                )
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + torch.stack([attn(x) for attn in self.attn_blocks]).sum(dim=0)
        x = x + torch.stack([mlp(x) for mlp in self.mlp_blocks]).sum(dim=0)

        return x


class Encoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        num_parallel: int,
        dropout: float,
        attention_dropout: float,
        dpr: list[float],
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        layers = []
        if dropout > 0.0:
            layers.append(nn.Dropout(dropout))

        for i in range(num_layers):
            layers.append(
                EncoderParallelBlock(
                    num_heads,
                    hidden_dim,
                    mlp_dim,
                    num_parallel,
                    dropout,
                    attention_dropout,
                    dpr[i],
                    activation_layer=nn.GELU,
                    layer_scale_init_value=layer_scale_init_value,
                    norm_layer=norm_layer,
                )
            )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # torch._assert(x.dim() == 3, f"Expected (batch_size, seq_length, hidden_dim) got {x.size()}")
        x = self.block(x)

        return x

    def forward_features(self, x: torch.Tensor) -> list[torch.Tensor]:
        xs = []
        for blk in self.block:
            x = blk(x)
            xs.append(x)

        return xs


# pylint: disable=invalid-name,too-many-instance-attributes
class ViT_Parallel(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    block_group_regex = r"encoder\.block\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        num_parallel: int = self.config["num_parallel"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        class_token: bool = self.config.get("class_token", True)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.hidden_dim = hidden_dim
        self.layer_scale_init_value = layer_scale_init_value
        self.num_reg_tokens = num_reg_tokens
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=True,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.num_special_tokens = 0

        # Add a class token
        if class_token is True:
            self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
            seq_length += 1
            self.num_special_tokens += 1
        else:
            self.class_token = None

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            seq_length += self.num_reg_tokens
            self.num_special_tokens += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))  # from BERT

        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            num_parallel,
            dropout,
            attention_dropout,
            dpr,
            layer_scale_init_value,
            norm_layer=norm_layer,
        )
        self.norm = norm_layer(hidden_dim, eps=1e-6)

        self.return_stages = ["neck"]  # Actually meaningless, but for completeness
        self.return_channels = [hidden_dim]
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            EncoderParallelBlock,
            16,
            mlp_dim=None,
            num_parallel=num_parallel,
            dropout=0,
            attention_dropout=0,
            drop_path=0,
            activation_layer=nn.GELU,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
        )

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embedding

        if H == self.size[0] and W == self.size[1]:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // self.patch_size, W // self.patch_size),
            self.num_special_tokens,
            antialias=False,
        )

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (H, W) = x.shape[-2:]
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        x = x + self._get_pos_embed(H, W)
        x = self.encoder(x)
        x = self.norm(x)

        x = x[:, self.num_special_tokens :]
        x = x.permute(0, 2, 1)
        (B, C, _) = x.size()
        x = x.reshape(B, C, H // self.patch_size, W // self.patch_size)

        return {self.return_stages[0]: x}

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad = False

        self.pos_embedding.requires_grad = False

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        x = x + pos_embedding[:, self.num_special_tokens :, :]

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

        # Append class and register tokens
        if self.class_token is not None:
            cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
            batch_class_token = cls_token.expand(x.shape[0], -1, -1)
            x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            if self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        (H, W) = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        x = x + self._get_pos_embed(H, W)
        x = self.encoder(x)
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            (B, C, _) = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        x = x + self._get_pos_embed(H, W)
        x = self.encoder(x)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)

        if self.class_token is None:
            x = x[:, self.num_special_tokens :]
            return x.mean(dim=1)

        # Classifier "token" as used by standard language architectures
        return x[:, self.num_reg_tokens]

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        # Add back class tokens
        self.pos_embedding = nn.Parameter(
            # On rounding error see: https://github.com/facebookresearch/dino/issues/8
            adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                self.num_special_tokens,
            )
        )


registry.register_model_config(
    "vit_parallel_s16_18x2_ls",
    ViT_Parallel,
    config={
        "patch_size": 16,
        "num_layers": 18,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_parallel": 2,
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "vit_parallel_s16_18x2_ls_avg",
    ViT_Parallel,
    config={
        "patch_size": 16,
        "num_layers": 18,
        "num_heads": 6,
        "hidden_dim": 384,
        "mlp_dim": 1536,
        "num_parallel": 2,
        "layer_scale_init_value": 1e-5,
        "class_token": False,
        "drop_path_rate": 0.05,
    },
)
registry.register_model_config(
    "vit_parallel_b16_18x2_ls",
    ViT_Parallel,
    config={
        "patch_size": 16,
        "num_layers": 18,
        "num_heads": 12,
        "hidden_dim": 768,
        "mlp_dim": 3072,
        "num_parallel": 2,
        "layer_scale_init_value": 1e-5,
        "drop_path_rate": 0.2,
    },
)



import json
import logging
import unittest

import torch
from parameterized import parameterized

from birder.data.collators.detection import batch_images
from birder.model_registry import registry
from birder.net.detection import base

logging.disable(logging.CRITICAL)


class TestBase(unittest.TestCase):
    def test_get_signature(self) -> None:
        signature = base.get_detection_signature((1, 3, 224, 224), 10, dynamic=False)
        self.assertIn("dynamic", signature)
        self.assertIn("inputs", signature)
        self.assertIn("outputs", signature)
        self.assertIn("boxes", signature["outputs"][0][0])


class TestNetDetection(unittest.TestCase):
    @parameterized.expand(  # type: ignore[misc]
        [
            ("deformable_detr", "fasternet_t0"),
            ("deformable_detr", "efficientvit_msft_m0"),  # 3 stage network
            ("deformable_detr_boxref", "regnet_x_200m"),
            ("detr", "regnet_y_1_6g"),
            ("efficientdet_d0", "efficientnet_v1_b0"),
            ("faster_rcnn", "resnet_v2_18"),
            ("faster_rcnn", "efficientvit_msft_m0"),  # 3 stage network
            ("fcos", "tiny_vit_5m"),
            ("fcos", "vit_s32"),  # 1 stage network
            ("retinanet", "mobilenet_v3_small_1_0"),
            ("retinanet", "efficientvit_msft_m0"),  # 3 stage network
            ("retinanet_sfp", "vit_det_m16_rms"),
            ("ssd", "efficientnet_v2_s"),
            ("ssd", "vit_s16"),  # 1 stage network
            ("ssdlite", "mobilenet_v2_0_25", (512, 512)),
            ("ssdlite", "vit_t16"),  # 1 stage network
            ("vitdet", "vit_sam_b16"),
        ]
    )
    def test_net_detection(
        self,
        network_name: str,
        encoder: str,
        size: tuple[int, int] = (256, 256),
    ) -> None:
        backbone = registry.net_factory(encoder, 3, 10, size=size)
        n = registry.detection_net_factory(network_name, 10, backbone, size=size, export_mode=True)

        # Ensure config is serializable
        _ = json.dumps(n.config)

        # Test network
        n.eval()
        out = n(torch.rand((1, 3, *size)))
        (detections, losses) = out
        self.assertEqual(len(losses), 0)
        for detection in detections:
            for key in ["boxes", "labels", "scores"]:
                self.assertFalse(torch.isnan(detection[key]).any())

        # Again in "dynamic size" mode
        (images, masks, image_sizes) = batch_images(
            [torch.rand((3, *size)), torch.rand((3, size[0] - 12, size[1] - 24))], size_divisible=4
        )
        out = n(images, masks=masks, image_sizes=image_sizes)

        # Reset classifier
        n.reset_classifier(20)
        n(torch.rand((1, 3, *size)))

        n.train()
        out = n(
            torch.rand((1, 3, *size)),
            targets=[
                {
                    "boxes": torch.tensor([[10.1, 10.1, 30.2, 40.2]]),
                    "labels": torch.tensor([1]),
                }
            ],
        )
        (detections, losses) = out
        self.assertGreater(len(losses), 0)
        for loss in losses.values():
            self.assertFalse(torch.isnan(loss).any())

        loss = sum(v for v in losses.values())
        self.assertEqual(loss.ndim, 0)

        for detection in detections:
            for key in ["boxes", "labels", "scores"]:
                self.assertFalse(torch.isnan(detection[key]).any())

        if n.scriptable is True:
            torch.jit.script(n)
        else:
            n.eval()
            torch.jit.trace(n, example_inputs=torch.rand((1, 3, *size)))
            n.train()

        # Freeze
        n.eval()
        n.freeze(freeze_classifier=False)
        n(torch.rand((1, 3, *size)))



import argparse
from typing import Any

import torch
from rich.columns import Columns
from rich.console import Console

from birder.common import cli
from birder.common import fs_ops
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import SignatureType
from birder.net.detection.base import DetectionSignatureType


def get_model_info(net: torch.nn.Module) -> dict[str, float]:
    num_params = 0
    param_size = 0
    buffer_size = 0
    for param in net.parameters():
        num_params += param.numel()
        param_size += param.numel() * param.element_size()

    for buffer in net.buffers():
        buffer_size += buffer.numel() * buffer.element_size()

    return {"num_params": num_params, "model_size": param_size + buffer_size}


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "model-info",
        allow_abbrev=False,
        help="print information about the model",
        description="print information about the model",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools model-info -n deit_b16 -t intermediate -e 0\n"
            "python -m birder.tools model-info --network squeezenet --epoch 100\n"
            "python -m birder.tools model-info --network densenet_121 -e 100 --pt2\n"
            "python -m birder.tools model-info -n efficientnet_v2_m -e 200 --lite\n"
            "python -m birder.tools model-info --network faster_rcnn --backbone resnext_101 -e 0\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to load (i.e. resnet_v2_50)"
    )
    subparser.add_argument("--backbone", type=str, help="the neural network to used as backbone")
    subparser.add_argument("--backbone-tag", type=str, help="backbone training log tag (loading only)")
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("--pts", default=False, action="store_true", help="load torchscript network")
    subparser.add_argument("--pt2", default=False, action="store_true", help="load standardized model")
    subparser.add_argument("--st", "--safetensors", default=False, action="store_true", help="load Safetensors weights")
    subparser.add_argument("--classes", default=False, action="store_true", help="print all classes")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.backbone is not None and registry.exists(args.backbone, net_type=DetectorBackbone) is False:
        raise cli.ValidationError(
            f"--backbone {args.network} not supported, see list-models tool for available options"
        )

    # Load model
    device = torch.device("cpu")
    signature: SignatureType | DetectionSignatureType
    backbone_custom_config = None
    if args.backbone is None:
        (net, (class_to_idx, signature, rgb_stats, custom_config)) = fs_ops.load_model(
            device,
            args.network,
            tag=args.tag,
            epoch=args.epoch,
            inference=True,
            reparameterized=args.reparameterized,
            pts=args.pts,
            pt2=args.pt2,
            st=args.st,
        )

    else:
        (net, (class_to_idx, signature, rgb_stats, custom_config, backbone_custom_config)) = (
            fs_ops.load_detection_model(
                device,
                args.network,
                tag=args.tag,
                backbone=args.backbone,
                backbone_tag=args.backbone_tag,
                epoch=args.epoch,
                inference=True,
                pts=args.pts,
                pt2=args.pt2,
                st=args.st,
            )
        )

    model_info = get_model_info(net)
    is_nan = torch.stack([torch.isnan(p).any() for p in net.parameters()]).any().item()

    console = Console()
    console.print(f"Network type: [bold]{type(net).__name__}[/bold], with task={net.task}")
    console.print(f"Network signature: {signature}")
    console.print(f"Network rgb values: {rgb_stats}")
    if custom_config is not None:
        console.print(f"Network has saved custom config: {custom_config}")
    if backbone_custom_config is not None:
        console.print(f"Network backbone has saved custom config: {backbone_custom_config}")

    console.print(f"Number of parameters: {model_info['num_params']:,}")
    console.print(f"Model size (inc. buffers): {(model_info['model_size']) / 1024**2:,.2f} [bold]MB[/bold]")
    console.print()
    if args.classes is True:
        console.print(Columns(list(class_to_idx.keys()), column_first=True, title="[bold]Class list[/bold]"))
    if is_nan is True:
        console.print()
        console.print("[red]Warning, NaN detected at the model weights[/red]")



import json
import logging
from collections import Counter
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypedDict

import polars as pl
import torch
from rich.console import Console
from rich.table import Table
from rich.text import Text
from torchmetrics.detection.mean_ap import MeanAveragePrecision

from birder.conf import settings

logger = logging.getLogger(__name__)

MetricsType = TypedDict(
    "MetricsType",
    {
        "map": float,
        "map_50": float,
        "map_75": float,
        "map_small": float,
        "map_medium": float,
        "map_large": float,
        "mar_1": float,
        "mar_10": float,
        "mar_100": float,
        "mar_small": float,
        "mar_medium": float,
        "mar_large": float,
        "ious": dict[tuple[int, int], torch.Tensor],
        "precision": torch.Tensor,
        "recall": torch.Tensor,
        "scores": torch.Tensor,
        "map_per_class": list[float],
        "mar_100_per_class": list[float],
        "classes": list[int],
    },
)


class Results:
    """
    Detection result analysis class
    """

    def __init__(
        self,
        sample_paths: list[str],
        targets: list[dict[str, Any]],
        detections: list[dict[str, torch.Tensor]],
        class_to_idx: dict[str, int],
    ):
        assert len(sample_paths) == len(targets)
        assert len(sample_paths) == len(detections)

        detections = [{k: v.cpu() for k, v in detection.items()} for detection in detections]
        targets = [{k: v.cpu() if isinstance(v, torch.Tensor) else v for k, v in t.items()} for t in targets]
        for target in targets:
            if "image_id" in target:
                del target["image_id"]

            # TorchMetrics can't handle "empty" images
            if "boxes" not in target:
                target["boxes"] = torch.tensor([], dtype=torch.float, device=torch.device("cpu"))
                target["labels"] = torch.tensor([], dtype=torch.int64, device=torch.device("cpu"))

        metrics = MeanAveragePrecision(
            iou_type="bbox", box_format="xyxy", class_metrics=True, extended_summary=True, average="macro"
        )
        metrics(detections, targets)
        metrics_dict = metrics.compute()

        self._iou_thresholds = metrics.iou_thresholds
        self._class_to_idx = class_to_idx
        self._label_names = ["Background"] + list(class_to_idx.keys())
        self._detections = detections
        self._targets = targets
        self._sample_paths = sample_paths

        self.metrics_dict: MetricsType = {
            "map": metrics_dict["map"].item(),
            "map_50": metrics_dict["map_50"].item(),
            "map_75": metrics_dict["map_75"].item(),
            "map_small": metrics_dict["map_small"].item(),
            "map_medium": metrics_dict["map_medium"].item(),
            "map_large": metrics_dict["map_large"].item(),
            "mar_1": metrics_dict["mar_1"].item(),
            "mar_10": metrics_dict["mar_10"].item(),
            "mar_100": metrics_dict["mar_100"].item(),
            "mar_small": metrics_dict["mar_small"].item(),
            "mar_medium": metrics_dict["mar_medium"].item(),
            "mar_large": metrics_dict["mar_large"].item(),
            "ious": metrics_dict["ious"],
            "precision": metrics_dict["precision"],
            "recall": metrics_dict["recall"],
            "scores": metrics_dict["scores"],
            "map_per_class": metrics_dict["map_per_class"].tolist(),
            "mar_100_per_class": metrics_dict["mar_100_per_class"].tolist(),
            "classes": metrics_dict["classes"].tolist(),
        }

    def __len__(self) -> int:
        return len(self._sample_paths)

    def __repr__(self) -> str:
        head = self.__class__.__name__
        body = [
            f"Number of samples: {len(self)}",
        ]
        body.append(f"mAP: {self.map:.4f}")

        lines = [head] + ["    " + line for line in body]

        return "\n".join(lines)

    @property
    def map(self) -> float:
        return self.metrics_dict["map"]

    def detailed_report(self) -> pl.DataFrame:
        """
        Returns a detailed detection report with per-class metrics
        """

        object_count: Counter[int] = Counter()
        for t in self._targets:
            object_count.update(t["labels"].tolist())

        row_list = []
        for class_num, mean_ap in zip(self.metrics_dict["classes"], self.metrics_dict["map_per_class"]):
            if mean_ap < 0:
                continue

            row_list.append(
                {
                    "Class": class_num,
                    "Class name": self._label_names[class_num],
                    "mAP": mean_ap,
                    "Objects": object_count[class_num],
                }
            )

        return pl.DataFrame(row_list)

    def log_short_report(self) -> None:
        """
        Log using the Python logging module a short metrics summary
        """

        report_df = self.detailed_report()
        total_objects = report_df["Objects"].sum()
        lowest_map = report_df[report_df["mAP"].arg_min()]  # type: ignore[index]
        highest_map = report_df[report_df["mAP"].arg_max()]  # type: ignore[index]

        logger.info(f"mAP {self.map:.4f} on {len(self)} images with {total_objects} objects")
        logger.info(f"Lowest mAP {lowest_map['mAP'][0]:.4f} for '{lowest_map['Class name'][0]}'")
        logger.info(f"Highest mAP {highest_map['mAP'][0]:.4f} for '{highest_map['Class name'][0]}'")

    def pretty_print(
        self,
        sort_by: Literal["class", "map"] = "class",
        order: Literal["ascending", "descending"] = "ascending",
        n: Optional[int] = None,
    ) -> None:
        console = Console()

        table = Table(show_header=True, header_style="bold dark_magenta")
        table.add_column("Class")
        table.add_column("Class name", style="dim")
        table.add_column("mAP", justify="right")
        table.add_column("Objects", justify="right")

        report_df = self.detailed_report()
        total_objects = report_df["Objects"].sum()
        if sort_by == "map":
            sort_column = "mAP"
        else:
            sort_column = sort_by.capitalize()

        report_df = report_df.sort(sort_column, descending=order == "descending")
        if n is not None:
            report_df = report_df[:n]

        for row in report_df.iter_rows(named=True):
            map_msg = f"{row['mAP']:.4f}"
            if row["mAP"] < 0.4:
                map_msg = "[red1]" + map_msg + "[/red1]"
            elif row["mAP"] < 0.5:
                map_msg = "[dark_orange]" + map_msg + "[/dark_orange]"

            table.add_row(
                f"{row['Class']}",
                row["Class name"],
                map_msg,
                f"{row['Objects']}",
            )

        console.print(table)

        map_text = Text()
        map_text.append(f"mAP {self.map:.4f} on {len(self)} images with {total_objects} objects")

        console.print(map_text)

    def save(self, name: str) -> None:
        """
        Save results object to file

        Parameters
        ----------
        name
            output file name.
        """

        detections = [{k: v.numpy().tolist() for k, v in detection.items()} for detection in self._detections]
        targets = [{k: v.numpy().tolist() for k, v in target.items()} for target in self._targets]
        output = dict(zip(self._sample_paths, detections))
        output["targets"] = dict(zip(self._sample_paths, targets))
        output["class_to_idx"] = self._class_to_idx

        if settings.RESULTS_DIR.exists() is False:
            logger.info(f"Creating {settings.RESULTS_DIR} directory...")
            settings.RESULTS_DIR.mkdir(parents=True)

        results_path = settings.RESULTS_DIR.joinpath(name)
        logger.info(f"Saving results at {results_path}")

        with open(results_path, "w", encoding="utf-8") as handle:
            json.dump(output, handle, indent=2)

    @staticmethod
    def load(path: str) -> "Results":
        """
        Load results object from file

        Parameters
        ----------
        path
            path to load from.
        """

        # Read label names
        with open(path, "r", encoding="utf-8") as handle:
            data: dict[str, Any] = json.load(handle)

        targets = data.pop("targets")
        class_to_idx = data.pop("class_to_idx")

        sample_paths = list(data.keys())
        detections = [{k: torch.tensor(v) for k, v in detection.items()} for detection in data.values()]
        targets = [{k: torch.tensor(v) for k, v in target.items()} for target in targets.values()]

        return Results(sample_paths, targets, detections, class_to_idx=class_to_idx)



import logging
from collections.abc import Callable
from typing import Any
from typing import Literal
from typing import NotRequired
from typing import Optional
from typing import TypedDict

import torch
import torch.nn.functional as F
from torch import nn

from birder.model_registry import Task
from birder.model_registry import registry

logger = logging.getLogger(__name__)

DataShapeType = TypedDict("DataShapeType", {"data_shape": list[int]})
SignatureType = TypedDict("SignatureType", {"inputs": list[DataShapeType], "outputs": list[DataShapeType]})
TokenOmissionResultType = TypedDict(
    "TokenOmissionResultType",
    {"tokens": NotRequired[torch.Tensor], "embedding": NotRequired[torch.Tensor]},
)
TokenRetentionResultType = TypedDict(
    "TokenRetentionResultType",
    {"features": NotRequired[torch.Tensor], "embedding": NotRequired[torch.Tensor]},
)


def get_signature(input_shape: tuple[int, ...], num_outputs: int) -> SignatureType:
    return {
        "inputs": [{"data_shape": [0, *input_shape[1:]]}],
        "outputs": [{"data_shape": [0, num_outputs]}],
    }


def make_divisible(v: float, divisor: int, min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original TensorFlow repository.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """

    if min_value is None:
        min_value = divisor

    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)

    # Make sure that round down does not go down by more than 10%
    if new_v < 0.9 * v:
        new_v += divisor

    return new_v


# class MiscNet(nn.Module):
#     """
#     Base class for general-purpose neural networks with automatic model registration

#     MiscNet provides a minimal foundation for integrating arbitrary PyTorch models into a unified model ecosystem.
#     Unlike specialized base classes (e.g., DetectionBaseNet for object detection),
#     MiscNet imposes minimal constraints and is suitable for any neural network architecture
#     that doesn't fit into specific task categories.
#     """

#     auto_register = False
#     scriptable = True
#     task = str(Task.MISCELLANEOUS)

#     def __init_subclass__(cls) -> None:
#         if cls.auto_register is False:
#             # Exclude networks with custom config (initialized only with aliases)
#             return

#         registry.register_model(cls.__name__.lower(), cls)

#     def __init__(self, *, config: Optional[dict[str, Any]] = None) -> None:
#         super().__init__()
#         if hasattr(self, "config") is False:  # Avoid overriding aliases
#             self.config = config
#         elif config is not None:
#             assert self.config is not None
#             self.config.update(config)  # Override with custom config

#     def forward(self, x: torch.Tensor) -> torch.Tensor:
#         raise NotImplementedError


class BaseNet(nn.Module):
    default_size: tuple[int, int] = (224, 224)
    block_group_regex: Optional[str]
    auto_register = False
    scriptable = True
    square_only = False
    task = str(Task.IMAGE_CLASSIFICATION)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls.__name__ in ["PreTrainEncoder", "DetectorBackbone"]:
            # Exclude all other base classes here
            return

        if cls.auto_register is False:
            # Exclude networks with custom config (initialized only with aliases)
            return

        registry.register_model(cls.__name__.lower(), cls)

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__()
        self.input_channels = input_channels
        self.num_classes = num_classes
        if hasattr(self, "config") is False:  # Avoid overriding aliases
            self.config = config
        elif config is not None:
            assert self.config is not None
            self.config.update(config)  # Override with custom config

        if size is not None:
            self.size = size
        else:
            self.size = self.default_size

        assert isinstance(self.size, tuple)
        assert isinstance(self.size[0], int)
        assert isinstance(self.size[1], int)
        if self.square_only is True:
            assert self.size[0] == self.size[1]

        self.dynamic_size = False

        self.classifier: nn.Module
        self.embedding_size: int

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Linear(embed_dim, self.num_classes)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.classifier = self.create_classifier()

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        logger.debug(f"Setting dynamic size to: {dynamic_size}")
        self.dynamic_size = dynamic_size

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        """
        Override this when one time adjustments for different resolutions is required.
        This should run after load_state_dict.
        """
        if new_size == self.size:
            return

        if self.square_only is True:
            assert new_size[0] == new_size[1]

        logger.info(f"Adjusting model input resolution from {self.size} to {new_size}")
        self.size = new_size

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True
        if unfreeze_features is True and hasattr(self, "features") is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        """
        Return full feature map, including special tokens
        """

        raise NotImplementedError

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        return self.classifier(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embedding(x)
        return self.classify(x)


class PreTrainEncoder(BaseNet):  # pylint: disable=abstract-method
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        self.max_stride: int = 32
        self.stem_stride: int
        self.stem_width: int
        self.encoding_size: int
        self.decoder_block: Callable[[int], nn.Module]


class MaskedTokenOmissionMixin:
    num_special_tokens: int

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        # Returned size (N, L, D)
        raise NotImplementedError


class MaskedTokenRetentionMixin:
    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        # Returned features size (B, C, H, W), embedding (B, D)
        raise NotImplementedError


class DetectorBackbone(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        self.return_stages = ["stage1", "stage2", "stage3", "stage4"]
        self.return_channels: list[int]

    def transform_to_backbone(self) -> None:
        if hasattr(self, "features") is True:
            self.features = nn.Identity()  # pylint: disable=attribute-defined-outside-init

        self.classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        raise NotImplementedError

    def freeze_stages(self, up_to_stage: int) -> None:
        raise NotImplementedError


def pos_embedding_sin_cos_2d(
    h: int, w: int, dim: int, num_special_tokens: int, temperature: int = 10000
) -> torch.Tensor:
    assert (dim % 4) == 0, "feature dimension must be multiple of 4 for sin-cos emb"

    (y, x) = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
    omega = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature**omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]
    pe = torch.concat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)

    if num_special_tokens > 0:
        pe = torch.concat([torch.zeros([num_special_tokens, dim]), pe], dim=0)

    return pe


def interpolate_attention_bias(
    attention_bias: torch.Tensor,
    old_resolution: tuple[int, int],
    new_resolution: tuple[int, int],
    mode: Literal["bilinear", "bicubic"] = "bicubic",
) -> torch.Tensor:
    (H, _) = attention_bias.size()

    # Interpolate
    orig_dtype = attention_bias.dtype
    attention_bias = attention_bias.float()  # Interpolate needs float32
    attention_bias = attention_bias.reshape(1, old_resolution[0], old_resolution[1], H).permute(0, 3, 1, 2)
    attention_bias = F.interpolate(attention_bias, size=new_resolution, mode=mode, antialias=True)
    attention_bias = attention_bias.permute(0, 2, 3, 1).reshape(H, new_resolution[0] * new_resolution[1])
    attention_bias = attention_bias.to(orig_dtype)

    return attention_bias


def reparameterize_available(net: nn.Module) -> bool:
    return hasattr(net, "reparameterize_model")



import argparse
import logging
import time
from itertools import combinations
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import polars as pl
import torch
from PIL import Image
from scipy.spatial import distance_matrix
from scipy.spatial.distance import pdist
from scipy.spatial.distance import squareform
from sklearn.manifold import TSNE
from torch.utils.data import DataLoader
from tqdm import tqdm

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.data.datasets.directory import make_image_dataset
from birder.data.transforms.classification import inference_preset
from birder.inference import classification

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals
def similarity(args: argparse.Namespace) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    logger.info(f"Using device {device}")

    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_model(
        device, args.network, tag=args.tag, epoch=args.epoch, inference=True, reparameterized=args.reparameterized
    )

    size = lib.get_size_from_signature(signature)
    batch_size = 32
    dataset = make_image_dataset(args.data_path, class_to_idx, transforms=inference_preset(size, rgb_stats, 1.0))
    inference_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4,
    )

    embedding_list: list[npt.NDArray[np.float32]] = []
    sample_paths: list[str] = []
    tic = time.time()
    with tqdm(total=len(dataset), initial=0, unit="images", unit_scale=True, leave=False) as progress:
        for file_paths, inputs, _targets in inference_loader:
            # Predict
            inputs = inputs.to(device)
            (_out, embedding) = classification.infer_batch(net, inputs, return_embedding=True)
            embedding_list.append(embedding)  # type: ignore[arg-type]
            sample_paths.extend(file_paths)

            # Update progress bar
            progress.update(n=batch_size)

    embeddings = np.concatenate(embedding_list, axis=0)

    toc = time.time()
    rate = len(dataset) / (toc - tic)
    (minutes, seconds) = divmod(toc - tic, 60)
    logger.info(f"{int(minutes):0>2}m{seconds:04.1f}s to classify {len(dataset)} samples ({rate:.2f} samples/sec)")

    logger.info("Processing similarity...")

    if args.cosine is True:
        distance_arr = pdist(embeddings, metric="cosine")

    else:
        # Dimensionality reduction
        tsne_embeddings_arr = TSNE(
            n_components=4, method="exact", learning_rate="auto", init="random", perplexity=20
        ).fit_transform(embeddings)

        # Build distance data frame
        distance_arr = distance_matrix(tsne_embeddings_arr, tsne_embeddings_arr)
        distance_arr = squareform(distance_arr)

    (sample_1, sample_2) = list(zip(*combinations(sample_paths, 2)))
    distance_df = pl.DataFrame(
        {
            "sample_1": sample_1,
            "sample_2": sample_2,
            "distance": distance_arr,
        }
    )
    distance_df = distance_df.sort("distance", descending=args.reverse)

    # Show image pairs
    if args.limit is None:
        args.limit = len(distance_df)

    for idx, pair in enumerate(distance_df[: args.limit].iter_rows(named=True)):
        (fig, (ax1, ax2)) = plt.subplots(2, 1)
        ax1.imshow(Image.open(pair["sample_1"]))
        ax1.set_title(pair["sample_1"])
        ax2.imshow(Image.open(pair["sample_2"]))
        ax2.set_title(pair["sample_2"])
        logger.info(f"{pair['distance']:.3f} distance between {pair['sample_1']} and {pair['sample_2']}")
        fig.suptitle(f"Distance = {pair['distance']:.3f} ({idx+1}/{len(distance_df):,})")
        plt.tight_layout()
        plt.show()


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "similarity",
        allow_abbrev=False,
        help="show images sorted by similarity",
        description="show images sorted by similarity",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools similarity -n efficientnet_v1_b4 -e 300 data/*/Alpine\\ swift\n"
            "python -m birder.tools similarity -n efficientnet_v2_s -e 200 --limit 3 data/*/Arabian\\ babbler\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to use (i.e. resnet_v2)"
    )
    subparser.add_argument("--cosine", default=False, action="store_true", help="use cosine distance")
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("--limit", type=int, help="limit number of pairs to show")
    subparser.add_argument("--reverse", default=False, action="store_true", help="start from most distinct pairs")
    subparser.add_argument("data_path", nargs="+", help="data files path (directories and files)")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    similarity(args)



import json
import logging
import unittest

import torch
from parameterized import parameterized

from birder.model_registry import registry
from birder.net.mim import base  # pylint: disable=unused-import # noqa: F401

logging.disable(logging.CRITICAL)


class TestNetMIM(unittest.TestCase):
    @parameterized.expand(  # type: ignore[misc]
        [
            ("crossmae", "deit3_t16"),
            ("crossmae", "deit3_reg4_t16"),
            ("crossmae", "rope_vit_b32"),
            ("crossmae", "rope_vit_reg4_b32"),
            ("crossmae", "rope_vit_so150m_p14_ap"),
            ("crossmae", "rope_vit_reg4_so150m_p14_ap"),
            ("crossmae", "simple_vit_b32"),
            ("crossmae", "vit_b32"),
            ("crossmae", "vit_reg4_b32"),
            ("crossmae", "vit_so150m_p14_ap"),
            ("crossmae", "vit_reg4_so150m_p14_ap"),
            ("crossmae", "vit_parallel_s16_18x2_ls"),
            ("fcmae", "convnext_v2_atto"),
            ("fcmae", "regnet_y_200m"),
            ("fcmae", "regnet_z_500m"),
            ("mae_hiera", "hiera_tiny"),
            ("mae_hiera", "hiera_abswin_tiny"),
            ("mae_vit", "deit3_t16"),
            ("mae_vit", "deit3_reg4_t16"),
            ("mae_vit", "rope_vit_b32"),
            ("mae_vit", "rope_vit_reg4_b32"),
            ("mae_vit", "rope_vit_so150m_p14_ap"),
            ("mae_vit", "rope_vit_reg4_so150m_p14_ap"),
            ("mae_vit", "simple_vit_b32"),
            ("mae_vit", "vit_b32"),
            ("mae_vit", "vit_reg4_b32"),
            ("mae_vit", "vit_so150m_p14_ap"),
            ("mae_vit", "vit_reg4_so150m_p14_ap"),
            ("mae_vit", "vit_parallel_s16_18x2_ls"),
            ("simmim", "hieradet_tiny"),
            ("simmim", "maxvit_t"),
            ("simmim", "nextvit_s"),
            ("simmim", "regnet_y_800m"),
            ("simmim", "swin_transformer_v2_t"),
            ("simmim", "swin_transformer_v2_w2_t"),
            ("simmim", "vit_b32"),
            ("simmim", "vit_reg4_b32"),
            ("simmim", "vit_so150m_p14_ap"),
            ("simmim", "vit_reg4_so150m_p14_ap"),
            ("simmim", "vit_parallel_s16_18x2_ls"),
        ]
    )
    def test_net_mim(self, network_name: str, encoder: str) -> None:
        encoder = registry.net_factory(encoder, 3, 10)
        size = (encoder.max_stride * 6, encoder.max_stride * 6)
        encoder.adjust_size(size)
        n = registry.mim_net_factory(network_name, encoder, size=size)

        # Ensure config is serializable
        _ = json.dumps(n.config)

        # Test network
        out = n(torch.rand((1, 3, *size)))
        for key in ["loss", "pred", "mask"]:
            self.assertFalse(torch.isnan(out[key]).any())

        self.assertEqual(out["loss"].ndim, 0)



"""
MaxViT, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/maxvit.py

Paper "MaxViT: Multi-Axis Vision Transformer", https://arxiv.org/abs/2204.01697
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from scipy import interpolate
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


def _get_conv_output_shape(
    input_size: tuple[int, int],
    kernel_size: tuple[int, int],
    stride: tuple[int, int],
    padding: tuple[int, int],
) -> tuple[int, int]:
    return (
        (input_size[0] - kernel_size[0] + 2 * padding[0]) // stride[0] + 1,
        (input_size[1] - kernel_size[1] + 2 * padding[1]) // stride[1] + 1,
    )


def _make_block_input_shapes(input_size: tuple[int, int], n_blocks: int) -> list[tuple[int, int]]:
    shapes = []
    block_input_shape = _get_conv_output_shape(input_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
    for _ in range(n_blocks):
        block_input_shape = _get_conv_output_shape(block_input_shape, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        shapes.append(block_input_shape)

    return shapes


def _get_relative_position_index(height: int, width: int) -> torch.Tensor:
    coords = torch.stack(torch.meshgrid([torch.arange(height), torch.arange(width)], indexing="ij"))
    coords_flat = torch.flatten(coords, 1)
    relative_coords = coords_flat[:, :, None] - coords_flat[:, None, :]
    relative_coords = relative_coords.permute(1, 2, 0).contiguous()
    relative_coords[:, :, 0] += height - 1
    relative_coords[:, :, 1] += width - 1
    relative_coords[:, :, 0] *= 2 * width - 1

    return relative_coords.sum(-1)


class MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expansion_ratio: float,
        squeeze_ratio: float,
        stride: tuple[int, int],
        activation_layer: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        p_stochastic_dropout: float,
    ) -> None:
        super().__init__()

        if stride[0] != 1 or stride[1] != 1 or in_channels != out_channels:
            self.proj = nn.Sequential(
                nn.AvgPool2d(kernel_size=(2, 2), stride=stride, padding=(0, 0)),
                nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True),
            )
        else:
            self.proj = nn.Identity()

        mid_channels = int(out_channels * expansion_ratio)
        sqz_channels = int(out_channels * squeeze_ratio)
        self.stochastic_depth = StochasticDepth(p_stochastic_dropout, mode="row")

        self.layers = nn.Sequential(
            norm_layer(in_channels),
            Conv2dNormActivation(
                in_channels,
                mid_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=activation_layer,
                norm_layer=norm_layer,
                inplace=None,
            ),
            Conv2dNormActivation(
                mid_channels,
                mid_channels,
                kernel_size=(3, 3),
                stride=stride,
                padding=(1, 1),
                groups=mid_channels,
                bias=False,
                activation_layer=activation_layer,
                norm_layer=norm_layer,
                inplace=None,
            ),
            SqueezeExcitation(mid_channels, sqz_channels, activation=nn.SiLU),
            nn.Conv2d(
                in_channels=mid_channels,
                out_channels=out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=True,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = self.proj(x)
        x = self.stochastic_depth(self.layers(x))

        return res + x


class RelativePositionalMultiHeadAttention(nn.Module):
    def __init__(self, feat_dim: int, head_dim: int, h: int, w: int) -> None:
        super().__init__()

        if feat_dim % head_dim != 0:
            raise ValueError(f"feat_dim: {feat_dim} must be divisible by head_dim: {head_dim}")

        self.n_heads = feat_dim // head_dim
        self.head_dim = head_dim
        self.size = (h, w)
        self.max_seq_len = self.size[0] * self.size[1]

        self.to_qkv = nn.Linear(feat_dim, self.n_heads * self.head_dim * 3)
        self.scale_factor = feat_dim**-0.5

        self.merge = nn.Linear(self.head_dim * self.n_heads, feat_dim)
        self.relative_position_bias_table = nn.Parameter(
            torch.empty(((2 * self.size[0] - 1) * (2 * self.size[1] - 1), self.n_heads), dtype=torch.float32),
        )
        self.relative_position_index = nn.Buffer(_get_relative_position_index(self.size[0], self.size[1]))

        # Initialize with truncated normal the bias
        nn.init.trunc_normal_(self.relative_position_bias_table, std=0.02)

    def get_relative_positional_bias(self) -> torch.Tensor:
        bias_index = self.relative_position_index.view(-1)
        relative_bias = self.relative_position_bias_table[bias_index].view(self.max_seq_len, self.max_seq_len, -1)
        relative_bias = relative_bias.permute(2, 0, 1).contiguous()
        return relative_bias.unsqueeze(dim=0)

    # pylint: disable=invalid-name
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, G, P, D) = x.size()
        H = self.n_heads
        DH = self.head_dim

        qkv = self.to_qkv(x)
        (q, k, v) = torch.chunk(qkv, 3, dim=-1)

        q = q.reshape(B, G, P, H, DH).permute(0, 1, 3, 2, 4)
        k = k.reshape(B, G, P, H, DH).permute(0, 1, 3, 2, 4)
        v = v.reshape(B, G, P, H, DH).permute(0, 1, 3, 2, 4)

        k = k * self.scale_factor
        dot_prod = torch.einsum("B G H I D, B G H J D -> B G H I J", q, k)
        pos_bias = self.get_relative_positional_bias()

        dot_prod = F.softmax(dot_prod + pos_bias, dim=-1)

        out = torch.einsum("B G H I J, B G H J D -> B G H I D", dot_prod, v)
        out = out.permute(0, 1, 3, 2, 4).reshape(B, G, P, D)

        out = self.merge(out)
        return out


class SwapAxes(nn.Module):
    def __init__(self, axis0: int, axis1: int) -> None:
        super().__init__()
        self.axis0 = axis0
        self.axis1 = axis1

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = torch.swapaxes(x, self.axis0, self.axis1)
        return res


class WindowPartition(nn.Module):
    def forward(self, x: torch.Tensor, p: tuple[int, int]) -> torch.Tensor:
        (B, C, H, W) = x.size()
        (PH, PW) = p  # pylint: disable=invalid-name

        # Chunk up H and W dimensions
        x = x.reshape(B, C, H // PH, PH, W // PW, PW)
        x = x.permute(0, 2, 4, 3, 5, 1)

        # Collapse P * P dimension
        x = x.reshape(B, (H // PH) * (W // PW), PH * PW, C)

        return x


class WindowDepartition(nn.Module):
    # pylint: disable=invalid-name
    def forward(self, x: torch.Tensor, p: tuple[int, int], h_partitions: int, w_partitions: int) -> torch.Tensor:
        (B, _G, _PP, C) = x.size()
        (PH, PW) = p  # pylint: disable=invalid-name
        HP = h_partitions
        WP = w_partitions

        # Split P * P dimension into 2 P tile dimension
        x = x.reshape(B, HP, WP, PH, PW, C)

        # Permute into B, C, HP, P, WP, P
        x = x.permute(0, 5, 1, 3, 2, 4)

        # Reshape into B, C, H, W
        x = x.reshape(B, C, HP * PH, WP * PW)

        return x


class PartitionAttentionLayer(nn.Module):
    """
    Layer for partitioning the input tensor into non-overlapping windows and applying attention to each window
    """

    def __init__(
        self,
        in_channels: int,
        head_dim: int,
        partition_size: tuple[int, int],
        partition_type: str,
        grid_size: tuple[int, int],
        mlp_ratio: int,
        activation_layer: Callable[..., nn.Module],
        norm_layer: Callable[..., nn.Module],
        attention_dropout: float,
        mlp_dropout: float,
        p_stochastic_dropout: float,
    ) -> None:
        super().__init__()

        self.n_heads = in_channels // head_dim
        self.head_dim = head_dim
        self.n_partitions = (grid_size[0] // partition_size[0], grid_size[1] // partition_size[1])
        self.partition_type = partition_type
        self.grid_size = grid_size

        if partition_type not in ["grid", "window"]:
            raise ValueError("partition_type must be either 'grid' or 'window'")

        if partition_type == "window":
            self.p = partition_size
            self.g = self.n_partitions
        else:
            self.p = self.n_partitions
            self.g = partition_size

        self.gh = self.grid_size[0] // self.p[0]
        self.gw = self.grid_size[1] // self.p[1]

        # Undefined behavior if H or W are not divisible by p
        # https://github.com/google-research/maxvit/blob/da76cf0d8a6ec668cc31b399c4126186da7da944/maxvit/models/maxvit.py#L766
        torch._assert(
            self.grid_size[0] % self.p[0] == 0 and self.grid_size[1] % self.p[1] == 0,
            f"Grid size must be divisible by partition size. Got grid size of "
            f"{self.grid_size} and partition size of {self.p}",
        )

        self.partition_op = WindowPartition()
        self.departition_op = WindowDepartition()
        self.partition_swap = SwapAxes(-2, -3) if partition_type == "grid" else nn.Identity()
        self.departition_swap = SwapAxes(-2, -3) if partition_type == "grid" else nn.Identity()

        self.attn_layer = nn.Sequential(
            norm_layer(in_channels),
            RelativePositionalMultiHeadAttention(in_channels, head_dim, partition_size[0], partition_size[1]),
            nn.Dropout(attention_dropout),
        )

        # Pre-normalization similar to transformer layers
        self.mlp_layer = nn.Sequential(
            nn.LayerNorm(in_channels),
            nn.Linear(in_channels, in_channels * mlp_ratio),
            activation_layer(),
            nn.Linear(in_channels * mlp_ratio, in_channels),
            nn.Dropout(mlp_dropout),
        )

        # layer scale factors
        self.stochastic_dropout = StochasticDepth(p_stochastic_dropout, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.partition_op(x, self.p)
        x = self.partition_swap(x)
        x = x + self.stochastic_dropout(self.attn_layer(x))
        x = x + self.stochastic_dropout(self.mlp_layer(x))
        x = self.departition_swap(x)
        x = self.departition_op(x, self.p, self.gh, self.gw)

        return x


class MaxVitLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        squeeze_ratio: float,
        expansion_ratio: float,
        stride: tuple[int, int],
        norm_layer: Callable[..., nn.Module],
        activation_layer: Callable[..., nn.Module],
        head_dim: int,
        mlp_ratio: int,
        mlp_dropout: float,
        attention_dropout: float,
        p_stochastic_dropout: float,
        partition_size: tuple[int, int],
        grid_size: tuple[int, int],
    ) -> None:
        super().__init__()
        self.layers = nn.Sequential(
            MBConv(
                in_channels=in_channels,
                out_channels=out_channels,
                expansion_ratio=expansion_ratio,
                squeeze_ratio=squeeze_ratio,
                stride=stride,
                activation_layer=activation_layer,
                norm_layer=norm_layer,
                p_stochastic_dropout=p_stochastic_dropout,
            ),
            PartitionAttentionLayer(
                in_channels=out_channels,
                head_dim=head_dim,
                partition_size=partition_size,
                partition_type="window",
                grid_size=grid_size,
                mlp_ratio=mlp_ratio,
                activation_layer=activation_layer,
                norm_layer=nn.LayerNorm,
                attention_dropout=attention_dropout,
                mlp_dropout=mlp_dropout,
                p_stochastic_dropout=p_stochastic_dropout,
            ),
            PartitionAttentionLayer(
                in_channels=out_channels,
                head_dim=head_dim,
                partition_size=partition_size,
                partition_type="grid",
                grid_size=grid_size,
                mlp_ratio=mlp_ratio,
                activation_layer=activation_layer,
                norm_layer=nn.LayerNorm,
                attention_dropout=attention_dropout,
                mlp_dropout=mlp_dropout,
                p_stochastic_dropout=p_stochastic_dropout,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.layers(x)
        return x


class MaxVitBlock(nn.Module):
    def __init__(
        self,
        # Conv parameters
        in_channels: int,
        out_channels: int,
        squeeze_ratio: float,
        expansion_ratio: float,
        # Conv + transformer parameters
        norm_layer: Callable[..., nn.Module],
        activation_layer: Callable[..., nn.Module],
        # Transformer parameters
        head_dim: int,
        mlp_ratio: int,
        mlp_dropout: float,
        attention_dropout: float,
        # Partitioning parameters
        partition_size: tuple[int, int],
        input_grid_size: tuple[int, int],
        # Number of layers
        n_layers: int,
        p_stochastic: list[float],
    ) -> None:
        super().__init__()
        if len(p_stochastic) != n_layers:
            raise ValueError(f"p_stochastic must have length n_layers={n_layers}, got p_stochastic={p_stochastic}.")

        # Account for the first stride of the first layer
        self.grid_size = _get_conv_output_shape(input_grid_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))

        layers = []
        for idx, p in enumerate(p_stochastic):
            if idx == 0:
                stride = (2, 2)
            else:
                stride = (1, 1)

            layers.append(
                MaxVitLayer(
                    in_channels=in_channels if idx == 0 else out_channels,
                    out_channels=out_channels,
                    squeeze_ratio=squeeze_ratio,
                    expansion_ratio=expansion_ratio,
                    stride=stride,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                    head_dim=head_dim,
                    mlp_ratio=mlp_ratio,
                    mlp_dropout=mlp_dropout,
                    attention_dropout=attention_dropout,
                    partition_size=partition_size,
                    grid_size=self.grid_size,
                    p_stochastic_dropout=p,
                ),
            )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.block(x)
        return x


class MaxViT(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.block\.(\d+)\.layers\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"
        assert self.size is not None, "must set size"

        image_size = self.size
        squeeze_ratio = 0.25
        expansion_ratio = 4
        mlp_ratio = 4
        mlp_dropout = 0.0
        attention_dropout = 0.0
        partition_size = (int(self.size[0] / (2**5)), int(self.size[1] / (2**5)))
        block_channels: list[int] = self.config["block_channels"]
        block_layers: list[int] = self.config["block_layers"]
        stem_channels: int = self.config["stem_channels"]
        head_dim: int = self.config["head_dim"]
        drop_path_rate: float = self.config["drop_path_rate"]

        # Make sure input size will be divisible by the partition size in all blocks
        # Undefined behavior if H or W are not divisible by p
        block_input_sizes = _make_block_input_shapes(image_size, len(block_channels))
        for idx, block_input_size in enumerate(block_input_sizes):
            if block_input_size[0] % partition_size[0] != 0 or block_input_size[1] % partition_size[1] != 0:
                raise ValueError(
                    f"Input size {block_input_size} of block {idx} is not divisible by partition size "
                    f"{partition_size}. Consider changing the partition size or the input size.\n"
                    f"Current configuration yields the following block input sizes: {block_input_sizes}"
                )

        norm_layer = partial(nn.BatchNorm2d, eps=1e-3, momentum=1 - 0.99)
        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                stem_channels,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                bias=False,
                norm_layer=norm_layer,
                activation_layer=nn.GELU,
                inplace=None,
            ),
            nn.Conv2d(
                stem_channels,
                stem_channels,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                bias=True,
            ),
        )

        # Account for stem stride
        image_size = _get_conv_output_shape(image_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.partition_size = partition_size

        # Blocks
        in_channels = [stem_channels] + block_channels[:-1]
        out_channels = block_channels

        # Pre-compute the stochastic depth probabilities from 0 to stochastic_depth_prob
        # since we have N blocks with L layers, we will have N * L probabilities uniformly distributed
        # over the range [0, stochastic_depth_prob]
        p_stochastic = torch.linspace(0, drop_path_rate, sum(block_layers)).tolist()

        p_idx = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, (in_channel, out_channel, num_layers) in enumerate(zip(in_channels, out_channels, block_layers)):
            stages[f"stage{i+1}"] = MaxVitBlock(
                in_channels=in_channel,
                out_channels=out_channel,
                squeeze_ratio=squeeze_ratio,
                expansion_ratio=expansion_ratio,
                norm_layer=norm_layer,
                activation_layer=nn.GELU,
                head_dim=head_dim,
                mlp_ratio=mlp_ratio,
                mlp_dropout=mlp_dropout,
                attention_dropout=attention_dropout,
                partition_size=partition_size,
                input_grid_size=image_size,
                n_layers=num_layers,
                p_stochastic=p_stochastic[p_idx : p_idx + num_layers],
            )
            return_channels.append(out_channel)
            image_size = stages[f"stage{i+1}"].grid_size
            p_idx += num_layers

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = block_channels[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 2
        self.stem_width = stem_channels
        self.encoding_size = block_channels[-1]

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Sequential(
            nn.LayerNorm(embed_dim),
            nn.Linear(embed_dim, embed_dim),
            nn.Tanh(),
            nn.Linear(embed_dim, self.num_classes, bias=False),
        )

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        new_grid_size = _get_conv_output_shape(new_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        new_grid_size = _get_conv_output_shape(new_grid_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        self.partition_size = (int(new_size[0] / (2**5)), int(new_size[1] / (2**5)))
        for m in self.body.modules():
            if isinstance(m, MaxVitBlock):
                m.grid_size = _get_conv_output_shape(new_grid_size, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
                for layer in m.block:
                    for i in range(1, 3):
                        mod = layer.layers[i]  # PartitionAttentionLayer
                        mod.n_partitions = (
                            new_grid_size[0] // self.partition_size[0],
                            new_grid_size[1] // self.partition_size[1],
                        )
                        mod.grid_size = new_grid_size

                        if mod.partition_type == "window":
                            mod.p = self.partition_size
                            mod.g = mod.n_partitions

                        else:
                            mod.p = mod.n_partitions
                            mod.g = self.partition_size

                        mod.gh = mod.grid_size[0] // mod.p[0]
                        mod.gw = mod.grid_size[1] // mod.p[1]

                        # Undefined behavior if H or W are not divisible by p
                        # https://github.com/google-research/maxvit/blob/da76cf0d8a6ec668cc31b399c4126186da7da944/maxvit/models/maxvit.py#L766
                        torch._assert(  # pylint: disable=protected-access
                            mod.grid_size[0] % mod.p[0] == 0 and mod.grid_size[1] % mod.p[1] == 0,
                            f"Grid size must be divisible by partition size. Got grid size of "
                            f"{mod.grid_size} and partition size of {mod.p}",
                        )

                        attn = mod.attn_layer[1]  # RelativePositionalMultiHeadAttention
                        old_attn_size = attn.size
                        attn.size = self.partition_size
                        attn.max_seq_len = self.partition_size[0] * self.partition_size[1]
                        attn.relative_position_index = nn.Buffer(
                            _get_relative_position_index(attn.size[0], attn.size[1])
                        )

                        # Interpolate relative_position_bias_table, adapted from
                        # https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/pos_embed_rel.py
                        dst_size = (2 * attn.size[0] - 1, 2 * attn.size[1] - 1)
                        rel_pos_bias = attn.relative_position_bias_table
                        rel_pos_bias = rel_pos_bias.detach()

                        num_attn_heads = rel_pos_bias.size(1)
                        src_size = (2 * old_attn_size[0] - 1, 2 * old_attn_size[1] - 1)

                        def _calc(src: int, dst: int) -> list[float]:
                            (left, right) = 1.01, 1.5
                            while right - left > 1e-6:
                                q = (left + right) / 2.0
                                gp = (1.0 - q ** (src // 2)) / (1.0 - q)  # Geometric progression
                                if gp > dst // 2:
                                    right = q

                                else:
                                    left = q

                            dis = []
                            cur = 1.0
                            for i in range(src // 2):
                                dis.append(cur)
                                cur += q ** (i + 1)

                            r_ids = [-_ for _ in reversed(dis)]
                            return r_ids + [0] + dis

                        y = _calc(src_size[0], dst_size[0])
                        x = _calc(src_size[1], dst_size[1])

                        ty = dst_size[0] // 2.0
                        tx = dst_size[1] // 2.0
                        dy = torch.arange(-ty, ty + 0.1, 1.0)
                        dx = torch.arange(-tx, tx + 0.1, 1.0)
                        dxy = torch.meshgrid(dx, dy, indexing="ij")

                        all_rel_pos_bias = []
                        for i in range(num_attn_heads):
                            z = rel_pos_bias[:, i].view(src_size[0], src_size[1]).float()
                            rgi = interpolate.RegularGridInterpolator(
                                (x, y), z.numpy().T, method="cubic", bounds_error=False, fill_value=None
                            )
                            r = torch.Tensor(rgi(dxy)).T.contiguous().to(rel_pos_bias.device)

                            r = r.view(-1, 1)
                            all_rel_pos_bias.append(r)

                        rel_pos_bias = torch.concat(all_rel_pos_bias, dim=-1)
                        attn.relative_position_bias_table = nn.Parameter(rel_pos_bias)

                new_grid_size = m.grid_size


registry.register_model_config(
    "maxvit_t",
    MaxViT,
    config={
        "block_channels": [64, 128, 256, 512],
        "block_layers": [2, 2, 5, 2],
        "stem_channels": 64,
        "head_dim": 32,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "maxvit_s",
    MaxViT,
    config={
        "block_channels": [96, 128, 256, 512],
        "block_layers": [2, 2, 5, 2],
        "stem_channels": 64,
        "head_dim": 32,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "maxvit_b",
    MaxViT,
    config={
        "block_channels": [96, 192, 384, 768],
        "block_layers": [2, 6, 14, 2],
        "stem_channels": 64,
        "head_dim": 32,
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "maxvit_l",
    MaxViT,
    config={
        "block_channels": [128, 256, 512, 1024],
        "block_layers": [2, 6, 14, 2],
        "stem_channels": 128,
        "head_dim": 32,
        "drop_path_rate": 0.5,
    },
)



import logging
import unittest

import numpy as np

from birder.conf import settings
from birder.results.classification import Results
from birder.results.classification import SparseResults
from birder.results.classification import top_k_accuracy_score

logging.disable(logging.CRITICAL)


class TestClassification(unittest.TestCase):
    def test_top_k_accuracy_score(self) -> None:
        y_true = np.array([0, 0, 2, 1, 1, 3])
        y_pred = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )
        indices = top_k_accuracy_score(y_true, y_pred, top_k=2)
        self.assertEqual(indices, [1, 2, 3])

    def test_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)
        self.assertFalse(results.missing_labels)
        self.assertAlmostEqual(results.accuracy, 1.0 / 6.0)
        self.assertAlmostEqual(results.top_k, 5.0 / 6.0)
        self.assertEqual(results._top_k_indices, [1, 2, 3, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

        cnf = results.confusion_matrix
        self.assertSequenceEqual(cnf.shape, (4, 4))

    def test_results_filter(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)
        filtered_results = results.filter_by_labels([0, 2])

        self.assertEqual(len(filtered_results), 3)
        self.assertFalse(filtered_results.missing_labels)
        self.assertAlmostEqual(filtered_results.accuracy, 1.0 / 3.0)

    def test_partial_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, settings.NO_LABEL, 2, settings.NO_LABEL, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)

        self.assertTrue(results.missing_labels)
        self.assertFalse(results.missing_all_labels)
        self.assertEqual(results._valid_length, 4)  # pylint: disable=protected-access
        self.assertAlmostEqual(results.accuracy, 1.0 / 4.0)
        self.assertAlmostEqual(results.top_k, 3.0 / 4.0)
        self.assertEqual(results._top_k_indices, [2, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

    def test_sparse_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)
        self.assertFalse(results.missing_labels)
        self.assertAlmostEqual(results.accuracy, 1.0 / 6.0)
        self.assertAlmostEqual(results.top_k, 5.0 / 6.0)
        self.assertEqual(results._top_k_indices, [1, 2, 3, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

        cnf = results.confusion_matrix
        self.assertSequenceEqual(cnf.shape, (4, 4))

    def test_sparse_results_filter(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)
        filtered_results = results.filter_by_labels([0, 2])

        self.assertEqual(len(filtered_results), 3)
        self.assertFalse(filtered_results.missing_labels)
        self.assertAlmostEqual(filtered_results.accuracy, 1.0 / 3.0)

    def test_partial_sparse_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, settings.NO_LABEL, 2, settings.NO_LABEL, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)

        self.assertTrue(results.missing_labels)
        self.assertFalse(results.missing_all_labels)
        self.assertEqual(results._valid_length, 4)  # pylint: disable=protected-access
        self.assertAlmostEqual(results.accuracy, 1.0 / 4.0)
        self.assertAlmostEqual(results.top_k, 3.0 / 4.0)
        self.assertEqual(results._top_k_indices, [2, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])



from birder.net.ssl.barlow_twins import BarlowTwins
from birder.net.ssl.byol import BYOL
from birder.net.ssl.capi import CAPIStudent
from birder.net.ssl.capi import CAPITeacher
from birder.net.ssl.data2vec import Data2Vec
from birder.net.ssl.data2vec2 import Data2Vec2
from birder.net.ssl.dino_v1 import DINO_v1
from birder.net.ssl.dino_v2 import DINOv2Student
from birder.net.ssl.dino_v2 import DINOv2Teacher
from birder.net.ssl.i_jepa import I_JEPA
from birder.net.ssl.ibot import iBOT
from birder.net.ssl.mmcr import MMCR
from birder.net.ssl.simclr import SimCLR
from birder.net.ssl.sscd import SSCD
from birder.net.ssl.vicreg import VICReg

__all__ = [
    "BarlowTwins",
    "BYOL",
    "CAPIStudent",
    "CAPITeacher",
    "Data2Vec",
    "Data2Vec2",
    "DINO_v1",
    "DINOv2Student",
    "DINOv2Teacher",
    "I_JEPA",
    "iBOT",
    "MMCR",
    "SimCLR",
    "SSCD",
    "VICReg",
]



import logging
import os
import unittest
import warnings

import requests

from birder.common.lib import get_pretrained_model_url
from birder.model_registry import registry
from birder.model_registry.model_registry import ModelRegistry
from birder.model_registry.model_registry import Task
from birder.model_registry.model_registry import group_sort
from birder.net.base import BaseNet
from birder.net.detection.base import DetectionBaseNet
from birder.net.mim.base import MIMBaseNet

logging.disable(logging.CRITICAL)


# pylint: disable=protected-access
class TestRegistry(unittest.TestCase):
    def test_group_sort(self) -> None:
        model_list = group_sort(
            ["b_prefix_b", "b_prefix_c", "a_prefix_z", "a_prefix_h", "a_prefix_a", "c_prefix_b", "c_prefix_a"]
        )
        self.assertEqual(
            model_list,
            ["a_prefix_z", "a_prefix_h", "a_prefix_a", "b_prefix_b", "b_prefix_c", "c_prefix_b", "c_prefix_a"],
        )

    def test_registry_nets(self) -> None:
        for net in registry._nets.values():
            self.assertTrue(issubclass(net, BaseNet))

        for net in registry._detection_nets.values():
            self.assertTrue(issubclass(net, DetectionBaseNet))

        for net in registry._mim_nets.values():
            self.assertTrue(issubclass(net, MIMBaseNet))

    def test_no_duplicates(self) -> None:
        all_names = []
        for net_name in registry._nets:
            all_names.append(net_name)

        for net_name in registry._detection_nets:
            all_names.append(net_name)

        for net_name in registry._mim_nets:
            all_names.append(net_name)

        self.assertEqual(len(all_names), len(set(all_names)))

    @unittest.skipUnless(os.environ.get("NETWORK_TESTS", False), "Avoid tests that require network access")
    def test_manifest(self) -> None:
        for model_name, model_metadata in registry._pretrained_nets.items():
            for model_format in model_metadata["formats"]:
                (_, url) = get_pretrained_model_url(model_name, model_format)

                try:
                    resp = requests.head(url, timeout=5, allow_redirects=True)
                except requests.RequestException:
                    # If first attempt fails with an exception, retry once
                    resp = requests.head(url, timeout=5, allow_redirects=True)

                with self.subTest(model_name=model_name, model_format=model_format):
                    self.assertEqual(resp.status_code, 200, f"{model_name} not found at {url}")
                    self.assertGreater(int(resp.headers["Content-Length"]), 100000)


# pylint: disable=protected-access
class TestModelRegistry(unittest.TestCase):
    def test_model_registry(self) -> None:
        model_registry = ModelRegistry()
        model_registry.register_model("net1", BaseNet)
        model_registry.register_model("net2", BaseNet)
        model_registry.register_model("net3", DetectionBaseNet)
        model_registry.register_model("net4", MIMBaseNet)

        self.assertListEqual(list(model_registry.all_nets.keys()), ["net1", "net2", "net3", "net4"])
        self.assertListEqual(list(model_registry._nets.keys()), ["net1", "net2"])
        self.assertListEqual(list(model_registry._detection_nets.keys()), ["net3"])
        self.assertListEqual(list(model_registry._mim_nets.keys()), ["net4"])
        self.assertListEqual(model_registry.list_models(task=Task.MASKED_IMAGE_MODELING), ["net4"])

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            model_registry.register_model("net1", BaseNet)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
